package cache

import (
	"errors"
	"fmt"
	"strings"
	"sync"

	corev1 "k8s.io/api/core/v1"
	k8scache "k8s.io/client-go/tools/cache"

	"gitlab.greeson.xyz/highsaltlevels/go-common/kubernetes"
	"gitlab.greeson.xyz/highsaltlevels/go-common/saltlog"
)

// Transformers intercept add, update, and delete operations to give consumers
// of the cache the ability to put their own custom item in the cache. If the
// transformer is not defined, the k8s resource is put in the cache instead.
// Return an error in the function if the operation should not occur (i.e. if
// the AddTransformer does not want the data added to the cache, it should
// return an error). Transformers MUST NOT panic.
type CacheTransformers struct {
	AddTransformer    func(key string, obj interface{}) (interface{}, error)
	UpdateTransformer func(key string, oldObj interface{}, newObj interface{}) (interface{}, error)
	DeleteTransformer func(key string, obj interface{}) error
}

// A ResourceCache implements Cache and stays up to date with objects
// via informers. Use NewResourceCache to instantiate this object.
type ResourceCache struct {
	Namespaces []string

	cacheType        int
	informers        map[string]k8scache.SharedIndexInformer
	items            map[string]interface{}
	itemsByNamespace map[string][]string
	lock             sync.Mutex
	stopCh           <-chan struct{}
	transformers     CacheTransformers
	logger           saltlog.Logger
}

// Get the item associated with the given Key. If the key does not exist,
// nil, and an ItemNotFoundError are returned.
func (r *ResourceCache) GetItem(key string) (interface{}, error) {
	r.lock.Lock()
	defer r.lock.Unlock()

	var item interface{}
	var exists bool
	if item, exists = r.items[key]; !exists {
		return nil, ItemNotFoundError
	}

	return item, nil
}

// Get all items in the cache that are in a particular namespace.
func (r *ResourceCache) GetItemsByNamespace(namespace string) ([]interface{}, error) {
	// Get the current items per index and use that everywhere in this
	// function to avoid race conditions. I don't want to use the lock
	// here because it's being used by the underlying GetItem call.
	itemsByNamespace, ok := r.itemsByNamespace[namespace]
	if !ok {
		return nil, ItemNotFoundError
	}

	items := make([]interface{}, 0)
	for _, key := range itemsByNamespace {
		item, err := r.GetItem(key)
		// It's possible the item can be deleted during this method. We should
		// just skip over those.
		if err != nil {
			if errors.Is(err, ItemNotFoundError) {
				continue
			}
		}

		items = append(items, item)
	}

	return items, nil
}

// Get all keys in the cache.
func (r *ResourceCache) ListKeys() ([]string, error) {
	r.lock.Lock()
	defer r.lock.Unlock()

	keys := make([]string, len(r.items))
	i := 0
	for key, _ := range r.items {
		keys[i] = key
		i++
	}

	return keys, nil
}

func (r *ResourceCache) addResource(obj interface{}) {
	var item interface{}
	key, namespace, err := r.getKeyAndNamespace(obj)
	if err != nil {
		r.logger.Warnf("skipping add due to: %v", err)
		return
	}

	// Use the UID of the resource as the key
	if r.transformers.AddTransformer != nil {
		item, err = r.transformers.AddTransformer(key, obj)
		if err != nil {
			r.logger.Warnf("skipping add due to: %v", err)
			return
		}
	}

	r.lock.Lock()
	defer r.lock.Unlock()
	r.items[key] = item
	r.itemsByNamespace[namespace] = append(r.itemsByNamespace[namespace], key)
}

func (r *ResourceCache) updateResource(oldObj interface{}, newObj interface{}) {
	var item interface{}
	key, _, err := r.getKeyAndNamespace(newObj)
	if err != nil {
		r.logger.Warnf("skipping update due to: %v", err)
		return
	}

	if r.transformers.UpdateTransformer != nil {
		item, err = r.transformers.UpdateTransformer(key, oldObj, newObj)
		if err != nil {
			r.logger.Warnf("skipping update due to: %v", err)
			return
		}
	}

	r.lock.Lock()
	defer r.lock.Unlock()
	// The itemsByNamespace should not change since nothing is added or
	// removed.
	r.items[key] = item
}

func (r *ResourceCache) deleteResource(obj interface{}) {
	key, namespace, err := r.getKeyAndNamespace(obj)
	if err != nil {
		r.logger.Warnf("skipping delete due to: %v", err)
		return
	}

	if r.transformers.DeleteTransformer != nil {
		err := r.transformers.DeleteTransformer(key, obj)
		if err != nil {
			r.logger.Warnf("skipping delete due to: %v", err)
			return
		}
	}

	r.lock.Lock()
	defer r.lock.Unlock()
	if _, exists := r.items[key]; exists {
		delete(r.items, key)
	} else {
		// This should never happen.
		r.logger.Errorf("no such key in cache: %s", key)
		return
	}

	keys := make([]string, 0)
	for _, searchKey := range r.itemsByNamespace[namespace] {
		if searchKey != key {
			keys = append(keys, searchKey)
		}
	}

	r.itemsByNamespace[namespace] = keys
}

func (r *ResourceCache) getKeyAndNamespace(obj interface{}) (key, namespace string, err error) {
	switch r.cacheType {
	case kubernetes.Pod:
		resource := obj.(*corev1.Pod)
		return string(resource.ObjectMeta.UID), resource.ObjectMeta.Namespace, nil

	case kubernetes.ConfigMap:
		resource := obj.(*corev1.ConfigMap)
		return string(resource.ObjectMeta.UID), resource.ObjectMeta.Namespace, nil

	default:
		// This should never happen
		return "", "", errors.New("no known cache type")
	}
}

// Create a new instance of a kubernetes resource cache. The underlying
// kubernetes client uses a service account token for authentication. If
// the runtime environment is not a kubernetes pod, NewResourceCache will
// attempt to use ${HOME}/.kube/config for authentication.
//
// Define CacheTransformers to act as interceptors for the underlying cache.
// If no CacheTransformers are defined, the kubernetes resource spec is
// stored in the cache. CacheTransformers are useful for storing smaller
// objects or appending custom metadata.
func NewResourceCache(resource string, namespaces []string, transformers CacheTransformers, logger saltlog.Logger) (*ResourceCache, error) {
	var resourceType int
	switch strings.ToUpper(resource) {
	case "POD":
		resourceType = kubernetes.Pod
	case "CONFIGMAP":
		resourceType = kubernetes.ConfigMap
	default:
		return nil, fmt.Errorf("unsupported resource type: %s", resource)

	}

	c := &ResourceCache{
		Namespaces: namespaces,

		items:            make(map[string]interface{}),
		itemsByNamespace: make(map[string][]string),
		cacheType:        resourceType,
		informers:        make(map[string]k8scache.SharedIndexInformer),
		lock:             sync.Mutex{},
		stopCh:           make(chan struct{}),
		transformers:     transformers,
		logger:           logger,
	}

	eh := k8scache.ResourceEventHandlerFuncs{
		AddFunc:    c.addResource,
		UpdateFunc: c.updateResource,
		DeleteFunc: c.deleteResource,
	}

	infs := make(map[string]k8scache.SharedIndexInformer, len(namespaces))
	for _, ns := range namespaces {
		inf, err := kubernetes.NewInformerFor(c.cacheType, ns)
		if err != nil {
			return nil, fmt.Errorf("failed to create informer for %s: %v", resource, err)
		}

		_, err = inf.AddEventHandler(eh)
		if err != nil {
			return nil, fmt.Errorf("failed to add event handler: %v", err)
		}

		c.logger.Infof("starting %s informer for namespace: %s", resource, ns)
		c.itemsByNamespace[ns] = make([]string, 0)
		infs[ns] = inf
		go inf.Run(c.stopCh)
	}

	for ns, inf := range infs {
		c.logger.Infof("waiting for %s informer to sync", ns)
		k8scache.WaitForCacheSync(c.stopCh, inf.HasSynced)
	}

	c.logger.Info("all informers have synced and cache is ready")
	return c, nil
}
