package cache

import (
	"errors"
	"strings"
	"sync"
	"testing"

	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	"gitlab.greeson.xyz/highsaltlevels/go-common/kubernetes"
	"gitlab.greeson.xyz/highsaltlevels/go-common/saltlog"
)

func mockAddTransformer(key string, obj interface{}) (interface{}, error) {
	return obj, nil
}

func mockAddTransformerError(key string, obj interface{}) (interface{}, error) {
	return nil, errors.New("expected error")
}

func mockUpdateTransformer(key string, oldObj, newObj interface{}) (interface{}, error) {
	return newObj, nil
}

func mockUpdateTransformerError(key string, oldObj, newObj interface{}) (interface{}, error) {
	return nil, errors.New("expected error")
}

func mockDeleteTransformer(key string, obj interface{}) error {
	return nil
}

func mockDeleteTransformerError(key string, obj interface{}) error {
	return errors.New("expected error")
}

func TestNewResourceCache(t *testing.T) {
	tests := []struct {
		name         string
		resource     string
		namespaces   []string
		transformers CacheTransformers

		err error
	}{
		{
			name:       "test success pod",
			resource:   "pod",
			namespaces: []string{"gitlab"},
		},
		{
			name:       "test success configmap",
			resource:   "cOnFiGmAp",
			namespaces: []string{"gitlab"},
		},
		{
			name:       "test unsupported resource type",
			resource:   "not supported",
			namespaces: []string{"gitlab"},
			err:        errors.New("unsupported resource type"),
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			_, err := NewResourceCache(tt.resource, tt.namespaces, tt.transformers, saltlog.NewLogger(saltlog.LevelDebug))
			switch err.(type) {
			case nil:
				if tt.err != nil {
					t.Fatalf("expected error: %v, but got no error", tt.err)
				}

			default:
				if tt.err == nil {
					t.Fatalf("expected no error, but got: %v", err)
				}

				if !strings.Contains(err.Error(), tt.err.Error()) {
					t.Fatalf("expected error: %v, but got: %v", tt.err, err)
				}

			}
		})
	}
}

func TestGetItem(t *testing.T) {
	tests := []struct {
		name string
		rc   *ResourceCache
		key  string
		item string
		err  error
	}{
		{
			name: "test get item successfully",
			rc: &ResourceCache{
				lock: sync.Mutex{},
				items: map[string]interface{}{
					"key": "value",
				},
			},
			key:  "key",
			item: "value",
		},
		{
			name: "test item not in cache",
			rc: &ResourceCache{
				lock: sync.Mutex{},
				items: map[string]interface{}{
					"differentkey": "value",
				},
			},
			key: "key",
			err: ItemNotFoundError,
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			item, err := tt.rc.GetItem(tt.key)
			switch err.(type) {
			case nil:
				if tt.err != nil {
					t.Fatalf("expected error: %v, but got no error", tt.err)
				}

				if item.(string) != tt.item {
					t.Errorf("expected to get: %v, but got: %v", tt.item, item)
				}
			default:
				if tt.err == nil {
					t.Fatalf("expected no error, but got: %v", err)
				}

				if !strings.Contains(err.Error(), tt.err.Error()) {
					t.Errorf("expected: %v, but got: %v", tt.err, err)
				}
			}
		})
	}
}

func TestGetItemsByNamespace(t *testing.T) {
	tests := []struct {
		name      string
		namespace string
		rc        *ResourceCache
		// Use a map for validation even though the return type is []string
		items map[string]string
		err   error
	}{
		{
			name:      "test get items successfully",
			namespace: "namespace",
			rc: &ResourceCache{
				lock: sync.Mutex{},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"key1", "key2"},
				},
				items: map[string]interface{}{
					"key1": "value1",
					"key2": "value2",
				},
			},
			items: map[string]string{
				"value1": "",
				"value2": "",
			},
		},
		{
			name:      "test namespace doesn't exist",
			namespace: "something else",
			rc: &ResourceCache{
				lock: sync.Mutex{},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"key1", "key2"},
				},
			},
			err: ItemNotFoundError,
		},
		{
			name:      "test item not in cache gets skipped",
			namespace: "namespace",
			rc: &ResourceCache{
				lock: sync.Mutex{},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"key1", "key2"},
				},
				items: map[string]interface{}{
					"key2": "value2",
				},
			},
			items: map[string]string{
				"value2": "",
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			items, err := tt.rc.GetItemsByNamespace(tt.namespace)
			switch err.(type) {
			case nil:
				if tt.err != nil {
					t.Fatalf("expected error: %v, but got no error", tt.err)
				}

				for _, value := range items {
					val := value.(string)
					if _, ok := tt.items[val]; !ok {
						t.Errorf("%v not in list of expected values %v", val, tt.items)
					}
				}
			default:
				if tt.err == nil {
					t.Fatalf("expected no error, but got: %v", err)
				}

				if !strings.Contains(err.Error(), tt.err.Error()) {
					t.Errorf("expected: %v, but got: %v", tt.err, err)
				}
			}
		})
	}
}

func TestListKeys(t *testing.T) {
	expectedKeys := map[string]interface{}{
		"key1": nil,
		"key2": nil,
	}
	rc := &ResourceCache{
		lock: sync.Mutex{},
		items: map[string]interface{}{
			"key1": "value1",
			"key2": "value2",
		},
	}

	keys, err := rc.ListKeys()

	if err != nil {
		t.Errorf("expected no error but got: %v", err)
	}

	for _, key := range keys {
		if _, ok := expectedKeys[key]; !ok {
			t.Errorf("%v not in list of expected values %v", key, expectedKeys)
		}
	}
}

func TestAddResource(t *testing.T) {
	tests := []struct {
		name        string
		obj         interface{}
		objKey      string
		namespace   string
		rc          *ResourceCache
		expectExist bool
	}{
		{
			name: "test add pod successfully",
			obj: &corev1.Pod{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: true,
			rc: &ResourceCache{
				cacheType:        kubernetes.Pod,
				lock:             sync.Mutex{},
				items:            make(map[string]interface{}),
				itemsByNamespace: make(map[string][]string),
			},
		},
		{
			name: "test add configmap successfully",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: true,
			rc: &ResourceCache{
				cacheType:        kubernetes.ConfigMap,
				lock:             sync.Mutex{},
				items:            make(map[string]interface{}),
				itemsByNamespace: make(map[string][]string),
			},
		},
		{
			name:        "test unknown cache type",
			obj:         &corev1.ConfigMap{},
			expectExist: false,
			rc: &ResourceCache{
				cacheType:        -1, // there should not be enough supported resources that iota will be negative
				lock:             sync.Mutex{},
				items:            make(map[string]interface{}),
				itemsByNamespace: make(map[string][]string),
				logger:           saltlog.NewLogger(saltlog.LevelDebug),
			},
		},
		{
			name: "test add configmap transformer",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: true,
			rc: &ResourceCache{
				cacheType:        kubernetes.ConfigMap,
				lock:             sync.Mutex{},
				items:            make(map[string]interface{}),
				itemsByNamespace: make(map[string][]string),
				transformers: CacheTransformers{
					AddTransformer: mockAddTransformer,
				},
			},
		},
		{
			name: "test add configmap transformer error",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: false,
			rc: &ResourceCache{
				cacheType:        kubernetes.ConfigMap,
				lock:             sync.Mutex{},
				items:            make(map[string]interface{}),
				itemsByNamespace: make(map[string][]string),
				transformers: CacheTransformers{
					AddTransformer: mockAddTransformerError,
				},
				logger: saltlog.NewLogger(saltlog.LevelDebug),
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tt.rc.addResource(tt.obj)
			if tt.expectExist {
				if _, ok := tt.rc.items[tt.objKey]; !ok {
					t.Errorf("%v not in list of expected values %v", tt.objKey, tt.rc.items)
				}

			} else {
				if _, ok := tt.rc.items[tt.objKey]; ok {
					t.Errorf("%v in cache when not expected", tt.objKey)
				}
			}
		})
	}
}

func TestUpdateResource(t *testing.T) {
	tests := []struct {
		name        string
		obj         interface{}
		objKey      string
		namespace   string
		rc          *ResourceCache
		expectExist bool
	}{
		{
			name: "test update pod successfully",
			obj: &corev1.Pod{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: true,
			rc: &ResourceCache{
				cacheType: kubernetes.Pod,
				lock:      sync.Mutex{},
				items: map[string]interface{}{
					"1234": "value",
				},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"1234"},
				},
			},
		},
		{
			name: "test update configmap successfully",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: true,
			rc: &ResourceCache{
				cacheType: kubernetes.ConfigMap,
				lock:      sync.Mutex{},
				items: map[string]interface{}{
					"1234": "value",
				},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"1234"},
				},
			},
		},
		{
			name:        "test unknown cache type",
			obj:         &corev1.ConfigMap{},
			expectExist: false,
			rc: &ResourceCache{
				cacheType:        -1,
				lock:             sync.Mutex{},
				items:            map[string]interface{}{},
				itemsByNamespace: map[string][]string{},
				logger:           saltlog.NewLogger(saltlog.LevelDebug),
			},
		},
		{
			name: "test update configmap transformer",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: true,
			rc: &ResourceCache{
				cacheType: kubernetes.ConfigMap,
				lock:      sync.Mutex{},
				items: map[string]interface{}{
					"1234": "value",
				},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"1234"},
				},
				transformers: CacheTransformers{
					UpdateTransformer: mockUpdateTransformer,
				},
			},
		},
		{
			name: "test update configmap transformer error",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: false,
			rc: &ResourceCache{
				cacheType:        kubernetes.ConfigMap,
				lock:             sync.Mutex{},
				items:            make(map[string]interface{}),
				itemsByNamespace: make(map[string][]string),
				transformers: CacheTransformers{
					UpdateTransformer: mockUpdateTransformerError,
				},
				logger: saltlog.NewLogger(saltlog.LevelDebug),
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tt.rc.updateResource(nil, tt.obj)
			if tt.expectExist {
				if _, ok := tt.rc.items[tt.objKey]; !ok {
					t.Errorf("%v not in list of expected values %v", tt.objKey, tt.rc.items)
				}

			} else {
				if _, ok := tt.rc.items[tt.objKey]; ok {
					t.Errorf("%v in cache when not expected", tt.objKey)
				}
			}
		})
	}
}

func TestDeleteResource(t *testing.T) {
	tests := []struct {
		name        string
		obj         interface{}
		objKey      string
		namespace   string
		rc          *ResourceCache
		expectExist bool
	}{
		{
			name: "test delete pod successfully",
			obj: &corev1.Pod{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: false,
			rc: &ResourceCache{
				cacheType: kubernetes.Pod,
				lock:      sync.Mutex{},
				items: map[string]interface{}{
					"1234": "value",
					"5678": "value2",
				},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"1234", "5678"},
				},
			},
		},
		{
			name: "test delete configmap successfully",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: false,
			rc: &ResourceCache{
				cacheType: kubernetes.ConfigMap,
				lock:      sync.Mutex{},
				items: map[string]interface{}{
					"1234": "value",
				},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"1234"},
				},
			},
		},
		{
			name:        "test unknown cache type",
			obj:         &corev1.ConfigMap{},
			expectExist: false,
			rc: &ResourceCache{
				cacheType: -1,
				lock:      sync.Mutex{},
				items: map[string]interface{}{
					"1234": "value",
				},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"1234"},
				},
				logger: saltlog.NewLogger(saltlog.LevelDebug),
			},
		},
		{
			name: "test delete configmap transformer",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: false,
			rc: &ResourceCache{
				cacheType: kubernetes.ConfigMap,
				lock:      sync.Mutex{},
				items: map[string]interface{}{
					"1234": "value",
				},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"1234"},
				},
				transformers: CacheTransformers{
					DeleteTransformer: mockDeleteTransformer,
				},
			},
		},
		{
			name: "test delete configmap transformer error",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: true,
			rc: &ResourceCache{
				cacheType: kubernetes.ConfigMap,
				lock:      sync.Mutex{},
				items: map[string]interface{}{
					"1234": "value",
				},
				itemsByNamespace: map[string][]string{
					"namespace": []string{"1234"},
				},
				transformers: CacheTransformers{
					DeleteTransformer: mockDeleteTransformerError,
				},
				logger: saltlog.NewLogger(saltlog.LevelDebug),
			},
		},
		{
			name: "test delete configmap item not in cache",
			obj: &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Namespace: "namespace",
					UID:       "1234",
				},
			},
			objKey:      "1234",
			namespace:   "namespace",
			expectExist: false,
			rc: &ResourceCache{
				cacheType:        kubernetes.ConfigMap,
				lock:             sync.Mutex{},
				items:            map[string]interface{}{},
				itemsByNamespace: map[string][]string{},
				logger:           saltlog.NewLogger(saltlog.LevelDebug),
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tt.rc.deleteResource(tt.obj)
			if tt.expectExist {
				if _, ok := tt.rc.items[tt.objKey]; !ok {
					t.Errorf("%v not in list of expected values %v", tt.objKey, tt.rc.items)
				}

			} else {
				if _, ok := tt.rc.items[tt.objKey]; ok {
					t.Errorf("%v in cache when not expected", tt.objKey)
				}
			}
		})
	}
}
