package handler

import (
	"io"
	"net/http"
	"os"

	"gitlab.greeson.xyz/highsaltlevels/go-common/saltlog"
)

type Handler interface {
	Handle(w http.ResponseWriter, r *http.Request)
}

// GenericFileHandler handles requests by serving the contents of the "path"
// that was specified in NewGenericFileHandler().
type GenericFileHandler struct {
	path        string
	contentType string
	data        []byte
}

// Handle sets the "Content-Type" header and writes the binary data from the file.
func (gfh GenericFileHandler) Handle(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Access-Control-Allow-Origin", "*")
	w.Header().Set("Content-Type", gfh.contentType)

	w.Write(gfh.data)
}

// NewGenericFileHandler creates a *GenericFileHandler object which loads the specified
// content upon instantiation. It will panic if it cannot load the content for any
// reason.
func NewGenericFileHandler(path, contentType string) *GenericFileHandler {
	data, err := os.ReadFile(path)
	if err != nil {
		saltlog.Panicf("genericfilehandler", "failed to read %s: %v", path, err)
	}

	return &GenericFileHandler{
		data:        data,
		contentType: contentType,
		path:        path,
	}
}

// PlainTextFileHandler handles requests by serving the contents of the "path"
// that was specified in NewPlainTextFileHandler.
type PlainTextFileHandler struct {
	path        string
	contentType string
	data        string
}

// Handle sets the "Content-Type" header and writes the string data from the file.
func (ptfh PlainTextFileHandler) Handle(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Access-Control-Allow-Origin", "*")
	w.Header().Set("Content-Type", ptfh.contentType)

	io.WriteString(w, ptfh.data)
}

// NewPlainTextFileHandler creates a *PlainTextFileHandler object which loads the
// specified content as plain text upon instantiation. It will panic if it cannot
// load the content for any reason.
func NewPlainTextFileHandler(path, contentType string) *PlainTextFileHandler {
	bytes, err := os.ReadFile(path)
	if err != nil {
		saltlog.Panicf("genericfilehandler", "failed to read %s: %v", path, err)
	}

	return &PlainTextFileHandler{
		data:        string(bytes[:]),
		contentType: contentType,
		path:        path,
	}
}

// HealthHandler handles requests by returning literally "ok\n".
type HealthHandler struct{}

// Handle sets the "Content-Type" header to "text/plain" and writes literally
// "ok\n".
func (hh HealthHandler) Handle(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Access-Control-Allow-Origin", "*")
	w.Header().Set("Content-Type", "text/plain")

	io.WriteString(w, "ok\n")
}

// NewHealthHandler creates a *HealthHandler object which only serves the string
// "ok\n".
func NewHealthHandler() *HealthHandler {
	return &HealthHandler{}
}

// CustomHandler handles requests by allowing an interceptor function to see the
// incoming request and form a custom response as a []byte.
type CustomHandler struct {
	handlerFunc func(w http.ResponseWriter, r *http.Request) []byte
	contentType string
}

// Handle sets the "Content-Type" header and writes whatever is returned by the
// call to the handlerFunc.
func (ch CustomHandler) Handle(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Access-Control-Allow-Origin", "*")
	w.Header().Set("Content-Type", ch.contentType)

	w.Write(ch.handlerFunc(w, r))
}

// NewCustomHandler creates a *CustomHandler object which stores a custom
// handler function to be called in the "Handle" method. The []byte returned
// by the handler function is what is written by the ResponseWriter.
func NewCustomHandler(handlerFunc func(w http.ResponseWriter, r *http.Request) []byte, contentType string) *CustomHandler {
	return &CustomHandler{
		handlerFunc: handlerFunc,
		contentType: contentType,
	}
}
