package handler

import (
	"errors"
	"net/http"
	"os"
	"strings"
	"testing"
)

// MockResponseWriter implements the http.ResponseWriter interface, but
// it just sets a boolean pointer instead of writing anything.
type MockResponseWriter struct {
	// Used to indicate if the writer was instructed to write or not.
	wrote *bool
}

func (m MockResponseWriter) Header() http.Header {
	return http.Header{}
}

func (m MockResponseWriter) Write([]byte) (int, error) {
	// Just set the boolean. Don't write anything.
	*m.wrote = true
	return 0, nil
}

func (m MockResponseWriter) WriteHeader(statusCode int) {}

func TestGenericFileHandler(t *testing.T) {
	// Use a mock response writer that doesn't write and sets a boolean
	// pointer instead. Since this is a pointer, to avoid nil panics, we
	// must instantiate a boolean variable.
	wrote := false
	w := &MockResponseWriter{wrote: &wrote}

	tests := []struct {
		name string
		path string
		err  error
	}{
		{
			name: "test read successfully",
			path: os.DevNull,
			err:  nil,
		},
		{
			name: "test file not found",
			path: "/this/path/should/not/exist/anywhere",
			err:  errors.New("no such file or directory"),
		},
	}

	for _, tt := range tests {
		wrote = false
		t.Run(tt.name, func(t *testing.T) {

			defer assertRecovery(t, tt.err)

			h := NewGenericFileHandler(tt.path, "content/type")
			h.Handle(w, nil)

			if !wrote {
				t.Errorf("mock response writer did not write when it was supposed to.")
			}
		})
	}
}

func TestPlainTextFileHandler(t *testing.T) {
	// Use a mock response writer that doesn't write and sets a boolean
	// pointer instead. Since this is a pointer, to avoid nil panics, we
	// must instantiate a boolean variable.
	wrote := false
	w := &MockResponseWriter{wrote: &wrote}

	tests := []struct {
		name string
		path string
		err  error
	}{
		{
			name: "test read successfully",
			path: os.DevNull,
			err:  nil,
		},
		{
			name: "test file not found",
			path: "/this/path/should/not/exist/anywhere",
			err:  errors.New("no such file or directory"),
		},
	}

	for _, tt := range tests {
		wrote = false
		t.Run(tt.name, func(t *testing.T) {

			defer assertRecovery(t, tt.err)

			h := NewPlainTextFileHandler(tt.path, "content/type")
			h.Handle(w, nil)

			if !wrote {
				t.Errorf("mock response writer did not write when it was supposed to.")
			}
		})
	}
}

func TestHealthHandler(t *testing.T) {
	// Use a mock response writer that doesn't write and sets a boolean
	// pointer instead. Since this is a pointer, to avoid nil panics, we
	// must instantiate a boolean variable.
	wrote := false
	w := &MockResponseWriter{wrote: &wrote}

	h := NewHealthHandler()
	h.Handle(w, nil)
	if !wrote {
		t.Errorf("mock response writer did not write when it was supposed to.")
	}
}

func TestCustomHandler(t *testing.T) {
	// Use a mock response writer that doesn't write and sets a boolean
	// pointer instead. Since this is a pointer, to avoid nil panics, we
	// must instantiate a boolean variable.
	wrote := false
	w := &MockResponseWriter{wrote: &wrote}

	var customHandlerFunc = func(w http.ResponseWriter, r *http.Request) []byte {
		return []byte("foo")
	}

	h := NewCustomHandler(customHandlerFunc, "content/type")
	h.Handle(w, nil)
	if !wrote {
		t.Errorf("mock response writer did not write when it was supposed to.")
	}
}

func assertRecovery(t *testing.T, expectedError error) {
	if rec := recover(); rec != nil {
		// If we get here, the function panicked
		if expectedError == nil {
			t.Fatalf("expected no panic but panicked with: %v", rec)
		}

		err := rec.(string)
		if !strings.Contains(err, expectedError.Error()) {
			t.Errorf("expected: %v, but got: %s", expectedError, err)
		}
	} else {
		if expectedError != nil {
			t.Fatalf("expected error: %v, but got no panic", expectedError)
		}
	}
}
