package kubernetes

import (
	"errors"
	"fmt"
	"path/filepath"
	"time"

	infcorev1 "k8s.io/client-go/informers/core/v1"
	kube "k8s.io/client-go/kubernetes"
	"k8s.io/client-go/rest"
	k8scache "k8s.io/client-go/tools/cache"
	"k8s.io/client-go/tools/clientcmd"
	"k8s.io/client-go/util/homedir"

	"gitlab.greeson.xyz/highsaltlevels/go-common/saltlog"
)

// These constants represent the supported kubernetes resouces
// to interface with.
const (
	Pod = iota
	ConfigMap
)

var client kube.Interface

// GetK8sClient returns a kubernetes client or an error. It attempts
// to use the in-cluster config, and if it can't, it tries to read
// ${HOME}/.kube/config. It returns an error if that fails.
func GetK8sClient() (kube.Interface, error) {
	var err error
	if client == nil {
		config, err := GetClientConfig()
		if err != nil {
			return nil, err
		}

		client, err = kube.NewForConfig(config)
	}

	return client, err
}

// NewInformerFor returns a SharedIndexInformer for a specified resource.
// See the constants of this package for a comprehensive list of supported
// resources. The resourceType parameter must be one of these constants.
func NewInformerFor(resourceType int, ns string) (k8scache.SharedIndexInformer, error) {
	client, err := GetK8sClient()
	if err != nil {
		return nil, fmt.Errorf("failed to get k8s client: %v", err)
	}

	switch resourceType {
	case Pod:
		return infcorev1.NewPodInformer(client, ns, time.Hour*24, nil), nil

	case ConfigMap:
		return infcorev1.NewConfigMapInformer(client, ns, time.Hour*24, nil), nil

	default:
		return nil, errors.New("unsupported resource type")
	}
}

func GetClientConfig() (config *rest.Config, err error) {
	// If we can't get the in-cluster config, try the ~/.kube/config
	if config, err = rest.InClusterConfig(); err == nil {
		return config, nil
	}

	l := saltlog.NewLogger(saltlog.LevelInfo)
	l.Warnf("could not initialize k8s client config: %v", err)
	l.Info("attempting to use ~/.kube/config")

	kubeconfig := filepath.Join(homedir.HomeDir(), ".kube", "config")
	config, err = clientcmd.BuildConfigFromFlags("", kubeconfig)
	return config, err
}
