package saltlog

import (
	"fmt"
	"runtime"
	"strings"
	"time"
)

// Default the printer to stdout, but allow tests to override it.
var printer = fmt.Print

// The Logger interface should support all log levels.
type Logger interface {
	Info(msg string)
	Debug(msg string)
	Warn(msg string)
	Error(msg string)
	Panic(msg string)
	Infof(msg string, args ...interface{})
	Debugf(msg string, args ...interface{})
	Warnf(msg string, args ...interface{})
	Errorf(msg string, args ...interface{})
	Panicf(msg string, args ...interface{})
}

// A SaltLogger is a simple logger that prints the component and time.
type SaltLogger struct {
	v int
	c string
}

// Enumeration type for log levels
const (
	LevelError = iota
	LevelWarning
	LevelInfo
	LevelDebug
)

// Info prints the time, component, and message as an info log to stdout.
func (l *SaltLogger) Info(msg string) {
	if l.v >= LevelInfo {
		printer(fmt.Sprintf("%s info  %s: %s\n", getTime(), l.c, msg))
	}
}

// Debug prints the time, component, and message as a debug log to stdout.
func (l *SaltLogger) Debug(msg string) {
	if l.v >= LevelDebug {
		printer(fmt.Sprintf("%s debug %s: %s\n", getTime(), l.c, msg))
	}
}

// Warn prints out the time, component, and message as a warn log to stdout.
func (l *SaltLogger) Warn(msg string) {
	if l.v >= LevelWarning {
		printer(fmt.Sprintf("%s warn  %s: %s\n", getTime(), l.c, msg))
	}
}

// Error prints out the time, component, and message as an error log to stdout.
func (l *SaltLogger) Error(msg string) {
	if l.v >= LevelError {
		printer(fmt.Sprintf("%s error %s: %s\n", getTime(), l.c, msg))
	}
}

// Panic prints out the time, component, and message but panics.
func (l *SaltLogger) Panic(msg string) {
	panic(fmt.Sprintf("%s panic %s: %s", getTime(), l.c, msg))
}

// Infof prints out the time, component, and message as an info log with formatted args.
func (l *SaltLogger) Infof(msg string, args ...interface{}) {
	if l.v >= LevelInfo {
		printer(fmt.Sprintf("%s info  %s: %s\n", getTime(), l.c, fmt.Sprintf(msg, args...)))
	}
}

// Debugf prints out the time, component, and message as a debug log with formatted args.
func (l *SaltLogger) Debugf(msg string, args ...interface{}) {
	if l.v >= LevelDebug {
		printer(fmt.Sprintf("%s debug %s: %s\n", getTime(), l.c, fmt.Sprintf(msg, args...)))
	}
}

// Warnf prints out the time, component, and message as a warn log with formatted args.
func (l *SaltLogger) Warnf(msg string, args ...interface{}) {
	if l.v >= LevelWarning {
		printer(fmt.Sprintf("%s warn  %s: %s\n", getTime(), l.c, fmt.Sprintf(msg, args...)))
	}
}

// Errorf prints out the time, component, and message as an error log with formatted args.
func (l *SaltLogger) Errorf(msg string, args ...interface{}) {
	if l.v >= LevelError {
		printer(fmt.Sprintf("%s error %s: %s\n", getTime(), l.c, fmt.Sprintf(msg, args...)))
	}
}

// Panicf prints out the time, component, and message with formatted args but panics.
func (l *SaltLogger) Panicf(msg string, args ...interface{}) {
	panic(fmt.Sprintf("%s panic %s: %s", getTime(), l.c, fmt.Sprintf(msg, args...)))
}

// NewLogger creates a simple Logger object that prints a formatted time,
// the log type, the component, and then the message. It supports rendering
// args like fmt.Sprintf.
func NewLogger(verbosity int) Logger {
	// Get the caller name.
	component := ""
	if _, file, _, ok := runtime.Caller(1); ok {
		files := strings.Split(file, "/")
		go_file_name := files[len(files)-1]
		component = strings.Split(go_file_name, ".")[0]
	}

	return &SaltLogger{
		c: component,
		v: verbosity,
	}
}

// Info creates a Logger and calls the Info method.
func Info(msg string) {
	NewLogger(LevelInfo).Info(msg)
}

// Debug creates a Logger and calls the Debug method.
func Debug(msg string) {
	NewLogger(LevelDebug).Debug(msg)
}

// Warn creates a Logger and calls the Warn method.
func Warn(msg string) {
	NewLogger(LevelWarning).Warn(msg)
}

// Error creates a Logger and calls the Error method.
func Error(msg string) {
	NewLogger(LevelError).Error(msg)
}

// Panic creates a Logger and calls the Panic method.
func Panic(msg string) {
	NewLogger(LevelDebug).Panic(msg)
}

// Infof creates a Logger and calls the Infof method.
func Infof(msg string, args ...interface{}) {
	NewLogger(LevelInfo).Infof(msg, args...)
}

// Debugf creates a Logger and calls the Debugf method.
func Debugf(msg string, args ...interface{}) {
	NewLogger(LevelDebug).Debugf(msg, args...)
}

// Warnf creates a Logger and calls the Warnf method.
func Warnf(msg string, args ...interface{}) {
	NewLogger(LevelWarning).Warnf(msg, args...)
}

// Errorf creates a Logger and calls the Errorf method.
func Errorf(msg string, args ...interface{}) {
	NewLogger(LevelError).Errorf(msg, args...)
}

// Panicf creates a Logger and calls the Panicf method.
func Panicf(msg string, args ...interface{}) {
	NewLogger(LevelDebug).Panicf(msg, args...)
}

func getTime() string {
	return time.Now().Format(time.DateTime)
}
