package saltlog

import (
	"fmt"
	"strings"
	"testing"
)

type mockPrinter struct {
	capturedLog string
}

func (m *mockPrinter) mockPrint(a ...any) (n int, err error) {
	if msg, ok := a[0].(string); ok {
		m.capturedLog = msg
		return 0, nil
	}

	panic("error in test, got non string argument to mockPrinter")
}

func TestSaltLogger(t *testing.T) {
	msg := "print me"

	// Set the printer to the mockPrinter
	mp := mockPrinter{}
	printer = mp.mockPrint

	tests := []struct {
		name    string
		logFunc func(string)
		message *string
	}{
		{
			name:    "test info log with info log level set",
			logFunc: Info,
			message: &msg,
		},
		{
			name:    "test debug log with debug log level set",
			logFunc: Debug,
			message: &msg,
		},
		{
			name:    "test warn log with warn log level set",
			logFunc: Warn,
			message: &msg,
		},
		{
			name:    "test error log with warn log level set",
			logFunc: Error,
			message: &msg,
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tt.logFunc(msg)

			if err := verifyLog(tt.message, mp.capturedLog); err != nil {
				t.Errorf("%v", err)
			}
		})
	}
}

func TestLoggerFormat(t *testing.T) {
	msg := "print me"

	// Set the printer to the mockPrinter
	mp := mockPrinter{}
	printer = mp.mockPrint

	tests := []struct {
		name    string
		logFunc func(string, ...interface{})
		message *string
	}{
		{
			name:    "test infof log with info log level set",
			logFunc: Infof,
			message: &msg,
		},
		{
			name:    "test debugf log with debug log level set",
			logFunc: Debugf,
			message: &msg,
		},
		{
			name:    "test warnf log with warn log level set",
			logFunc: Warnf,
			message: &msg,
		},
		{
			name:    "test errorf log with warn log level set",
			logFunc: Errorf,
			message: &msg,
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			tt.logFunc(msg)

			if err := verifyLog(tt.message, mp.capturedLog); err != nil {
				t.Errorf("%v", err)
			}
		})
	}
}

func TestLoggerPanic(t *testing.T) {
	expected := "testpanic"
	defer func() {
		err := recover()
		if err == nil {
			t.Fatalf("expected to panic but it did not")
		}

		actual, _ := err.(string)
		if !strings.Contains(actual, expected) {
			t.Errorf("expected: %s, but got: %s", expected, actual)
		}
	}()

	l := NewLogger(LevelDebug)
	l.Panic(expected)
}

func TestLoggerPanicf(t *testing.T) {
	expected := "test panic"
	defer func() {
		err := recover()
		if err == nil {
			t.Fatalf("expected to panic but it did not")
		}

		actual, _ := err.(string)
		if !strings.Contains(actual, expected) {
			t.Errorf("expected: %s, but got: %s", expected, actual)
		}
	}()

	l := NewLogger(LevelDebug)
	l.Panicf("test %s", "panic")
}

func TestPanic(t *testing.T) {
	expected := "expected"
	defer func() {
		err := recover()
		if err == nil {
			t.Fatalf("expected to panic but it did not")
		}

		actual, _ := err.(string)
		if !strings.Contains(actual, expected) {
			t.Errorf("expected: %s, but got: %s", expected, actual)
		}
	}()

	Panic(expected)
}

func TestPanicf(t *testing.T) {
	var expected1, expected2 = "expected1", "expected2"
	defer func() {
		err := recover()
		if err == nil {
			t.Fatalf("expected to panic but it did not")
		}

		actual, _ := err.(string)
		if !strings.Contains(actual, expected1) {
			t.Errorf("expected: %s, but got: %s", expected1, actual)
		}
		if !strings.Contains(actual, expected2) {
			t.Errorf("expected: %s, but got: %s", expected2, actual)
		}
	}()

	Panicf(expected1, "%s", expected2)
}

func verifyLog(expected *string, actual string) error {
	if expected != nil {
		if !strings.Contains(actual, *expected) {
			return fmt.Errorf("expected to print '%s' but printed: %s", *expected, actual)
		}
	} else {
		if actual != "" {
			return fmt.Errorf("expected not to print, but printed; %s", actual)
		}
	}
	return nil
}
